structure Check :>
sig
    datatype typOrError =
        WellTyped of Syntax.typ
      | IllTyped 
    
    val check : Syntax.exp -> typOrError
        
end  = 
struct

    open Syntax
    
    datatype typOrError =
        WellTyped of Syntax.typ
      | IllTyped 

    type context = (name * typ) list

    fun lookup (gamma : context, v : name) : typOrError = 
	case List.find (fn (v', t) => v = v') gamma of
            NONE => IllTyped
          | SOME (_, t) => WellTyped t

    (* Given a context gamma and expression e, return
       WellTyped tau if there is a tau such that gamma |- e : tau, or
       IllTyped otherwise *)
    fun checkOpen (g : context, e : exp) : typOrError = 
        case e of 
            Var v => lookup (g,v)
          | Zero => WellTyped Nat
          | Succ e => (case checkOpen (g, e) of
                          WellTyped Nat => WellTyped Nat
                        | _ => IllTyped)
          | Rec (e, e0, (x,y,e1)) => 
                (case checkOpen (g, e) of
                         WellTyped Nat =>
                             (case checkOpen(g, e0) of
                                  WellTyped t0 =>
                                      (case checkOpen ((x,Nat) :: (y,t0) :: g, e1) of
                                           WellTyped t1 => if t0 = t1 then WellTyped t0 else IllTyped
                                         | IllTyped => IllTyped)
                               |  IllTyped => IllTyped )
                       | _ => IllTyped)
          | Lam (t,(x,e)) => 
              (case checkOpen ((x,t) :: g, e) of
                   WellTyped t' => WellTyped (Arrow (t,t'))
                 | IllTyped => IllTyped)
          | App (e1,e2) => 
                (case checkOpen (g, e1) of
                     WellTyped (Arrow (t,t')) =>
                         (case checkOpen (g, e2) of
                              WellTyped t2 => if t = t2 then WellTyped t' else IllTyped 
                            | IllTyped => IllTyped)
                   | _ => IllTyped)
          | Let (e1, (x,e2)) => 
                     (case (checkOpen (g, e1)) of
                          WellTyped t1 => checkOpen (((x,t1) :: g), e2 )
                        | IllTyped => IllTyped)
          | Triv => raise Fail "unimplemented"
          | Pair(e1, e2) => raise Fail "unimplemented"
          | ProjLeft e => raise Fail "unimplemented"
          | ProjRight e => raise Fail "unimplemented"
          | InLeft (t2, e) => raise Fail "unimplemented"
          | InRight (t1, e) => raise Fail "unimplemented"
          | Case (e, (x1,e1), (x2, e2)) => raise Fail "unimplemented"
                    
    (* Given an expression e, 
       return WellTyped tau if there is a tau such that empty |- e : tau,
       or return IllTyped if e is not well-typed in the empty context *)
    fun check e = checkOpen ([], e)
	
end
