structure Step :>
sig
    datatype result =
        Done
      | Stepped of Syntax.exp

    (* Given a closed, well-typed expression e (i.e. empty |- e : tau for some tau), 
       returns an Stepped e' for an e' such that e |-> e' if such an e' exists, or 
       returns Done if e is done *)
    val progress : Syntax.exp -> result
        
end = 
struct

    open Syntax

    datatype result =
        Done
      | Stepped of Syntax.exp

    fun progress e = 
        case e of
            Zero => Done
          | Succ e =>
                (case progress e of
                     Done => Done
                   | Stepped e' => Stepped (Succ e'))
          | Rec (e, e0, (x,y,e1)) => 
                     Stepped (case progress e of
                                  Done => 
                                      (case e of
                                           Zero   => e0
                                         | Succ n' => (subst(subst (e1, (n',x)), (Rec (n', e0 , (x,y,e1)) , y))))
                                | Stepped e' => (Rec (e', e0, (x,y,e1))))
          | Lam (t,e) => Done
          | App (e1,e2) =>
                     Stepped (case progress e1 of
                                  Done => (case progress e2 of
                                               Done => (case e1 of Lam (t,(x,e)) => subst (e, (e2,x)))
                                             | Stepped e2' => App (e1 , e2'))
                                | Stepped e1' => App (e1', e2))
          | Let (e1,(x,e2)) => Stepped (case progress e1 of
                                            Done => (subst (e2, (e1,x)))
                                          | Stepped e1' => (Let (e1', (x,e2))))
          | Triv => raise Fail "unimplemented"
          | Pair(e1,e2) => raise Fail "unimplemented"
          | ProjLeft e => raise Fail "unimplemented"
          | ProjRight e =>  raise Fail "unimplemented"
          | InLeft (t,e) => raise Fail "unimplemented"
          | InRight (t,e) =>  raise Fail "unimplemented"
          | Case(e, (x1,e1), (x2,e2)) => raise Fail "unimplemented"
                                   
end
